/*
 * Copyright (C) 2009 Yoonsuck Choe
 * 
 * This file is an add-on of ANJI (Another NEAT Java Implementatioredn).
 * This add-on has the same terms and conditions as ANJI (see below).
 * This is a modified version of DoublePoleBalanceFitnessFunction.java
 * by Derek James.
 * 
 * Tue Nov  3 17:08:37 CST 2009
 *
 * --- original statement in DoublePoleBalanceFitnessFunction.java ---
 *
 * This file is a part of ANJI (Another NEAT Java Implementation).
 * ANJI is free software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; if
 * not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 * 
 * created by Derek James on Jul 5, 2005
 */

package com.anji.navi;

import java.awt.*;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.nio.DoubleBuffer;

import org.apache.log4j.Logger;
import org.jgap.BulkFitnessFunction;
import org.jgap.Chromosome;

import com.anji.imaging.IdentifyImageFitnessFunction;
import com.anji.integration.Activator;
import com.anji.integration.ActivatorTranscriber;
import com.anji.util.Arrays;
import com.anji.util.Configurable;
import com.anji.util.Properties;
import com.anji.util.Randomizer;
import com.anji.navi.TSDisplay;

/**
 * This is written from scratch by Yoonsuck Choe.
 * @author Yoonsuck Choe
 */
public class TSFitnessFunction implements BulkFitnessFunction, Configurable {

private final static String MAX_STEPS_KEY = "navi.maxsteps";
private final static String NUM_TRIALS_KEY = "navi.numtrials";
private final static String POP_SIZE_KEY = "popul.size";
private final static String UNIT_COUNT_KEY = "unit.count";
private final static String STIM_SIZE_KEY = "stimulus.size";
private final static String RESP_SIZE_KEY = "response.size";

// Some useful physical model constants.
private static final double DUMMY = -9.8;

// private PoleBalanceDisplay display = null;

private final static int DEFAULT_TIMESTEPS = 10000;
private int maxTimesteps = DEFAULT_TIMESTEPS;

private final static int DEFAULT_NUM_TRIALS = 10;
private int numTrials = DEFAULT_NUM_TRIALS;

private final static int DEFAULT_POP_SIZE = 150;
private int 	populationSize = DEFAULT_POP_SIZE;

private final static int DEFAULT_UNIT_COUNT = 5;
private int 	unitCount = DEFAULT_UNIT_COUNT;

private final static int DEFAULT_STIM_SIZE = 5;
private int 	stimSize = DEFAULT_STIM_SIZE;

private final static int DEFAULT_RESP_SIZE = 1;
private int 	respSize = DEFAULT_RESP_SIZE;

private double numSim = 20;

private boolean randomOpponentFlag = true;

private boolean dispFlag = false;

private final static Logger logger = Logger.getLogger( NaviFitnessFunction.class );

private ActivatorTranscriber factory;

private Random rand;

private int    debugLevel = 1;

private TSDisplay tsd;

private int 	dispCount = 0;
private int 	dispInterval = 10;

private double[][] effectMatrix = {
	{1.0, 0.3, 0.1, 1.0, 0.8},
	{1.0, 0.6, 0.2, 1.0, 1.0},
	{0.3, 1.0, 1.0, 0.2, 0.2},
	{0.1, 0.6, 0.9, 0.2, 0.1},
	{1.0, 0.6, 0.2, 1.0, 1.0}};

// private class to deal with individual agents
private class TSAgent {

	double x;
	double y;
	int type;
  	int typeSize = 5;
	double acc;	
	double r;
	double e;
	double res;
	double energyReductionRate = 0.7;
	double resourceReductionRate = 0.7;

	String tName[] = {"SA", "MG", "T", "R", "GL"};
	double tRange[] = {0.4, 0.6, 1.0, 0.6, 0.4};
	double tRadius[] = {0.1, 0,3, 1.0, 0.7, 0.5};
	double tAccuracy[] = {0.8, 0.9, 0.7, 0.4, 0.3};

	TSAgent (double x, double y, int t, double a, double r, double e, 
			double res) {
		this.x = x;
		this.y = y;
		this.type = t;
		this.acc = a;
		this.r = r;
		this.e = e;
		this.res = res;
	}

	TSAgent (double x, double y, int t, double e, double res) {
		this.x = x;
		this.y = y;
		this.type = t;
		this.acc = tAccuracy[t];
		this.r = tRadius[t];
		this.e = e;
		this.res = res;
		
	}

	public void setEnergy(double e) {
		this.e = e;
	} 
	
	public void setResource(double res) {
		this.res = res;
	} 
	
	public void reduceEnergy(double rate) {
		this.e = e * rate;
	} 
	
	public void reduceResource(double rate) {
		this.res = res * rate;
	} 
	
}

// Opposing teams
private TSAgent[] teamRed;
private TSAgent[] teamBlue;
private TSAgent[] teamRedCopy;
private TSAgent[] teamBlueCopy;

/**
 * @see com.anji.util.Configurable#init(com.anji.util.Properties)
 */
public void init( Properties props ) throws Exception {
	try {

		// 1. Init activator transcriber
		factory = (ActivatorTranscriber) props.singletonObjectProperty( ActivatorTranscriber.class );

		// 2. pull parameters from the properties file
		populationSize = props.getIntProperty(POP_SIZE_KEY, DEFAULT_POP_SIZE);
		maxTimesteps = props.getIntProperty(MAX_STEPS_KEY, DEFAULT_TIMESTEPS );
		numTrials = props.getIntProperty(NUM_TRIALS_KEY, DEFAULT_NUM_TRIALS );
		unitCount = props.getIntProperty(UNIT_COUNT_KEY, DEFAULT_UNIT_COUNT);
		stimSize = props.getIntProperty(STIM_SIZE_KEY, DEFAULT_STIM_SIZE);
		respSize = props.getIntProperty(RESP_SIZE_KEY, DEFAULT_RESP_SIZE);
			// * Other types: getBooleanProperty( INPUT_VELOCITY_KEY, true );

		// 3. set up randomizer  
		Randomizer randomizer = (Randomizer) props.singletonObjectProperty( Randomizer.class );
		rand = randomizer.getRand();

		// 4. set up display window
		if (dispFlag) {
			tsd = new TSDisplay(500,500);
		}

		// 5. set up initial scenario
	  	setupTeams();

	}
	catch ( Exception e ) {
		throw new IllegalArgumentException( "invalid properties: " + e.getClass().toString() + ": "
				+ e.getMessage() );
	}
}

public void debugMessage(int level, double[][] mat) {
	
    if (debugLevel>=level) {
	for (int i=0; i<mat.length; ++i) {
	System.out.print(java.util.Arrays.toString(mat[i])+"\n");
	}
	System.out.println("---");
    }

}

public void debugMessage(int level, int[] vec) {
	
    if (debugLevel>=level) {
	System.out.print(java.util.Arrays.toString(vec)+"\n");
    }

}

public void debugMessage(int level, double[] vec) {
	
    if (debugLevel>=level) {
	System.out.print(java.util.Arrays.toString(vec)+"\n");
    }

}

public void debugMessage(int level, String str) {
	
    if (debugLevel>=level) {
	System.out.print("DEBUG:"+str);
    }

}

public void debugMessageNoprefix(int level, String str) {
	
    if (debugLevel>=level) {
	System.out.print(str);
    }

}

/**
 * @see org.jgap.BulkFitnessFunction#evaluate(java.util.List)
 * @see IdentifyImageFitnessFunction#evaluate(Chromosome)
 * yschoe: this version allows competition among all pairs of peers to 
 * calculate the fitness.
 */
public void evaluate( List genotypes ) {

  if (randomOpponentFlag) {
  	evaluateListRand(genotypes);
  } else {
  	evaluateNtoN(genotypes);
  }

}

/**
 * make chromosome compete with random opponent
 */
public void evaluateListRand( List genotypes ) {

    	int result[] = new int [populationSize]; 

	Iterator it = genotypes.iterator();

	// 1. Run competition simulation, and repeat that for numTrials times.
	for (int i=0; i<numTrials; ++i) {
	  // 1.1 set up the teams
	  setupTeams();
	  backupTeams();
	
	  // 1.2 run a single trial
	  int id=0;
	  while (it.hasNext()) {
		Chromosome c = (Chromosome) it.next();
		result[id] += evaluateRandom(c);
		id++;
		restoreTeams();
	  } 

	  // reset iterator so that we can start anew.
	  it = genotypes.iterator(); 

	} 

	debugMessage(1,result);

	// 2. Set fitness values
	int id=0;
	it = genotypes.iterator();
	while (it.hasNext()) {
		Chromosome c = (Chromosome) it.next();
	  	c.setFitnessValue(result[id]+numTrials);
		id++;	
		restoreTeams();
	}


}

/**
 * make chromosome compete with all other opponents in the same generation
 * - needs to be fully implemented: evaluatePair()
 */
public void evaluateNtoN( List genotypes ) {

	debugMessage(1,"running evaluate list\n");

	//populationSize = genotypes.size();

    	int result = 0; // result of pair-wise competition
    	int fitness = 0; // overall fitness of one chromosome against all other.
	int resultTable[][] = new int [populationSize][populationSize];

	Iterator it = genotypes.iterator();

	// 1. Run pair-wise competition simulation, and repeat that for n times.
	for (int i=0; i<numTrials; ++i) {
	  // 1.1 set up the teams
	  //
	  setupTeams();
	  backupTeams();
	
	  int id1 = 0;

	  // 1.2 run a single trial
	  while (it.hasNext()) {
		Chromosome c1 = (Chromosome) it.next();
		Iterator opponent = genotypes.iterator();

		int id2=0;
		
		resultTable[id1][id1] = 0;
		// iterate over all other chromosomes in this generation
		// - alternative: consider comparing to random
		while (opponent.hasNext()) {
			restoreTeams();	
			Chromosome c2 = (Chromosome) opponent.next();
			// skip redundant pairs
			if (id2<=id1) {
				++id2;
				continue; 
			}
			result = evaluatePair(c1,c2);
			resultTable[id1][id2] = result;
			resultTable[id2][id1] = -result;
			++id2;
		}
		++id1;
	  } // end of single chromosome 

	  // reset iterator so that we can start anew.
	  it = genotypes.iterator(); 

	} // end of n trials loop

	// 2. calculate fitness
	it = genotypes.iterator();
	int id1 = 0;
	while (it.hasNext()) {
		Chromosome c = (Chromosome) it.next();
		Iterator opponent = genotypes.iterator();

		int resultSum=0;
		
		for (int id2 = 0; id2<populationSize; ++id2) {
			resultSum += resultTable[id1][id2];
		}			

		debugMessageNoprefix(3,(resultSum+populationSize-1)+"("
		 	+resultSum+")  ");
		c.setFitnessValue(resultSum+(populationSize-1)*numTrials);

		++id1;
	
	}

	debugMessageNoprefix(3,java.util.Arrays.deepToString(resultTable));

	debugMessage(3,"\n");

	dispCount++;
}

/**
 * Set up the two teams 
 */
public void setupTeams() {

	teamRed = new TSAgent[unitCount];
	teamBlue = new TSAgent[unitCount];

	teamRedCopy = new TSAgent[unitCount];
	teamBlueCopy = new TSAgent[unitCount];

	Color redColor[] = new Color[unitCount];
	Color blueColor[] = new Color[unitCount];

	for (int i=0; i<unitCount; ++i) {
		int step = 128/unitCount;
		redColor[i] = new Color(255,128+step*i,128+step*i);
		blueColor[i] = new Color(128+step*i,128+step*i,255);
	}

	debugMessage(3,effectMatrix);
	debugMessage(3," "+effectMatrix[3][2]+" ");

	// 1. setup location and energy
	if (dispFlag) {
		tsd.clearList();
	}
	double redE = 0.0;
	double blueE = 0.0;
	for (int i=0; i<unitCount; ++i) {
		int rType = rand.nextInt(5);
		teamRed[i] = new TSAgent(rand.nextDouble(),
			rand.nextDouble()/3.0+0.55,
			rType, rand.nextDouble()+0.5, 1.0);
		teamRedCopy[i] = new TSAgent(0,0,0,0,0);
		redE += teamRed[i].e;

		int bType = rand.nextInt(5);
		teamBlue[i] = new TSAgent(rand.nextDouble(),
			0.45-rand.nextDouble()/3.0,
			bType, rand.nextDouble()+0.5, 1.0);
		teamBlueCopy[i] = new TSAgent(0,0,0,0,0);
		blueE += teamBlue[i].e;

		//if (dispCount % dispInterval == 0) {
		if (dispFlag) {
		  tsd.addObj(teamRed[i].x,teamRed[i].y,redColor[rType], 
			teamRed[i].r, teamRed[i].e);
		  tsd.addObj(teamBlue[i].x,teamBlue[i].y,blueColor[bType], 
			teamBlue[i].r, teamBlue[i].e);
		}
	}

	// equalize total energy
	for (int i=0; i<unitCount; ++i) {
		teamBlue[i].e *= redE/blueE;
	}
	debugMessage(3,"Energy = "+redE+" "+blueE+"\n");

	// 2. setup dependency

}

public void restoreTeams() {
   for (int i=0 ; i<unitCount; ++i) {
	teamRed[i].x = teamRedCopy[i].x;
	teamRed[i].y = teamRedCopy[i].y;
	teamRed[i].type = teamRedCopy[i].type;
	teamRed[i].acc = teamRedCopy[i].acc;
	teamRed[i].r = teamRedCopy[i].r;
	teamRed[i].e = teamRedCopy[i].e;
	teamRed[i].res = teamRedCopy[i].res;

	teamBlue[i].x = teamBlueCopy[i].x;
	teamBlue[i].y = teamBlueCopy[i].y;
	teamBlue[i].type = teamBlueCopy[i].type;
	teamBlue[i].acc = teamBlueCopy[i].acc;
	teamBlue[i].r = teamBlueCopy[i].r;
	teamBlue[i].e = teamBlueCopy[i].e;
	teamBlue[i].res = teamBlueCopy[i].res;
   }
}

public void backupTeams() {
   for (int i=0 ; i<unitCount; ++i) {
	teamRedCopy[i].x = teamRed[i].x;
	teamRedCopy[i].y = teamRed[i].y;
	teamRedCopy[i].type = teamRed[i].type;
	teamRedCopy[i].acc = teamRed[i].acc;
	teamRedCopy[i].r = teamRed[i].r;
	teamRedCopy[i].e = teamRed[i].e;
	teamRedCopy[i].res = teamRed[i].res;

	teamBlueCopy[i].x = teamBlue[i].x;
	teamBlueCopy[i].y = teamBlue[i].y;
	teamBlueCopy[i].type = teamBlue[i].type;
	teamBlueCopy[i].acc = teamBlue[i].acc;
	teamBlueCopy[i].r = teamBlue[i].r;
	teamBlueCopy[i].e = teamBlue[i].e;
	teamBlueCopy[i].res = teamBlue[i].res;
   }
}

public void runSimOnce(int[] redOrder, int[] blueOrder) {

	// Red offensive
	double[] blueEnergy = new double[unitCount];

	double discountFactor = 0.9;

	for (int i=0; i<unitCount; ++i) {
		blueEnergy[i] = teamBlue[i].e;
	}
	
	for (int i=0; i<unitCount; ++i) {
		int target = redOrder[i];
		// effect
		double effect = teamRed[i].acc 
			* effectMatrix[teamRed[i].type][teamBlue[target].type];
		// only when enough resource
		if (teamRed[i].r > 0) {
			blueEnergy[target] *= (1.0-effect*discountFactor);
		}
	}

	// Blue offensive

	double[] redEnergy = new double[unitCount];

	for (int i=0; i<unitCount; ++i) {
		redEnergy[i] = teamRed[i].e;
	}
	
	for (int i=0; i<unitCount; ++i) {
		int target = blueOrder[i];
		// effect
		double effect = teamBlue[i].acc 
			* effectMatrix[teamBlue[i].type][teamRed[target].type];
		// only when enough resource
		if (teamBlue[i].r > 0) {
			redEnergy[target] *= (1.0-effect);
		}
	}

	for (int i=0; i<unitCount; ++i) {
		teamRed[i].e = redEnergy[i];
		teamRed[i].r -= 2.0/numSim;
		teamBlue[i].e = blueEnergy[i];
		teamBlue[i].r -= 2.0/numSim;
	}

}


public int analyzeSim() {

	double red=0;
	double blue=0;
	for (int i=0; i<unitCount; ++i) {
		red += teamRed[i].e;	
		blue += teamBlue[i].e;	
	}

	debugMessage(1,"result= "+red+" "+blue+"\n");

	if (red>blue) {
		debugMessageNoprefix(3,"1");
		return 1;
	} else {
		debugMessageNoprefix(3,"0");
		return -1;
	}
	//return rand.nextInt(3)-1;

}

/**
 * Evaluate pair of chromosomes against each other and set fitness for
 * the first chromosome.
 * @param c1, c2
 * returns 1 if c1 wins, -1 if c2 wins, 0 if error
 */
public int evaluatePair( Chromosome c1, Chromosome c2 ) {
	try {

		int[] redOrder=new int[unitCount];
                int[] blueOrder=new int[unitCount];

		for (int i=0; i<numSim; ++i) {
			redOrder = targetSelect(c1,teamRed,teamBlue);
			blueOrder = targetSelect(c2,teamBlue,teamRed);
			runSimOnce(redOrder,blueOrder);
		}
		
		debugMessage(3,redOrder);
		debugMessage(3,blueOrder);
		// 2. return game outcome. Based on this, fitness will be
		// calculated.
		return analyzeSim(); // or 1
	}
	catch ( Throwable e ) {
		logger.warn( "error evaluating chromosome "+c1.toString()
			+" and "+c2.toString(), e );
		return 0;
	}
}

/* 
 * target select using NEAT
 */
public int[] targetSelect( Chromosome c, TSAgent[] team, TSAgent[] team2 ) {

	try {
	int[] order = new int[unitCount];

	Activator activator = factory.newActivator( c );

	double[] input = new double[stimSize];
	double[] output = new double[respSize];
 	double result;

	debugMessage(3,"********\n");

	for (int i = 0; i<unitCount; ++i) {

		int maxIdx=0;
		double max=-9999;

		for (int j=0; j<unitCount; ++j) {
			// reset input
			for (int k=0; k<stimSize; ++k) { input[k] = 0.0; }
			// input 0 to 4
			//input[teamRed[i].type]=1.0;
			// input 10: range -- no effect yet
			input[0]= team[i].r;
			// input 11: distance -- no effect yet
			input[1] = 0.0; 
			// input 12: accuracy
			input[2] = team[i].acc;
			// input 13: resource
			input[3] = team[i].res;
			// input 14: nearby targets -- no effect yet
			input[4] = 0.0;
			// input 5 to 9
			//input[teamBlue[j].type+5]=1.0;
			// input 15: target energy -- no effect yet
			input[5] = 0.0; // teamBlue.e
			// input 17: effect
			input[6] = effectMatrix[team[i].type][team2[j].type];
			// input 17: bias
			input[7] = 1.0;

			// skip dead units
			if (team2[j].e < 0.01) {
				continue;
			}
			
			output = activator.next(input);	
			result = output[0];
			if (result > max) {
				max = result;
				maxIdx = j;
			}
			debugMessage(3,input);
		}

		order[i] = maxIdx;
	}

	debugMessage(3,order);

	return order;

	} catch ( Throwable e ) {
		int[] order = new int[unitCount];
		logger.warn( "error evaluating chromosome " + c.toString(), e );
		c.setFitnessValue( 0 );
		return order;
	}
}

/*
 * select target randomly
 */
public int[] targetSelectRand( ) {

	int[] order = new int[unitCount];

	for (int i=0; i<unitCount; ++i) {
		order[i] = rand.nextInt(unitCount);
	}

	debugMessage(3,order);

	return order;
}

public void plotLineOfSight ( int[] order1, int[] order2) {

	plotLineOfSight(order1,order2,255);

}

public void plotObj () {

  if (dispFlag) {
	Color redColor[] = new Color[unitCount];
	Color blueColor[] = new Color[unitCount];

	for (int i=0; i<unitCount; ++i) {
		int step = 128/unitCount;
		redColor[i] = new Color(255,128+step*i,128+step*i);
		blueColor[i] = new Color(128+step*i,128+step*i,255);
	}


	for (int i=0; i<unitCount; ++i) {
	 tsd.addObj(teamRed[i].x,teamRed[i].y,redColor[teamRed[i].type], 
		teamRed[i].r, teamRed[i].e);
	 tsd.addObj(teamBlue[i].x,teamBlue[i].y,blueColor[teamBlue[i].type],
		teamBlue[i].r,teamBlue[i].e);
	}
  }

}

public void plotLineOfSight ( int[] order1, int[] order2, int scale ) {

  if (dispFlag) {
	double x1, x2, y1, y2;
	Color red = new Color(scale,0,0);
	Color blue = new Color(0,0,scale);
	//tsd.clearLineList();
	for (int i=0; i<unitCount; ++i) { 
		x1 = teamRed[i].x;
		y1 = teamRed[i].y;
		x2 = teamBlue[order1[i]].x;
		y2 = teamBlue[order1[i]].y;
		tsd.addLine(x1,y1,x2,y2,red,7.0f);

		x1 = teamBlue[i].x;
		y1 = teamBlue[i].y;
		x2 = teamRed[order2[i]].x;
		y2 = teamRed[order2[i]].y;
		tsd.addLine(x1,y1,x2,y2,blue,2.0f);
	}
  }

}

/**
 * Evaluate chromosome against random opponent and set fitness for
 * the first chromosome.
 * @param c
 * returns 1 if c wins, -1 if c loses, 0 if error or draw
 */
public int evaluateRandom( Chromosome c ) {
	try {
		double curEnergySum = 1.0;
		double prevEnergySum = 0.0;
	
		int[] redOrder;
		int[] blueOrder = new int[unitCount];

		Random rand = new Random();

		// run sim until stablility is reached
		//while (curEnergySum - prevEnergySum > 0.01) {
		debugMessage(2,"****************\n");
		for (int i=0; i<numSim; ++i) {

		   // 1. Select target
		   
			// 1.1. RedTeam: NEAT
			// - each member target one opponent
			// - run each target through net and pick the
			//   highest ranking one first
			// - do basic sanity check: within range? etc.
			debugMessage(2,"["+i+"]");
			redOrder = targetSelect(c,teamRed,teamBlue);
			debugMessage(1,"red "+java.util.Arrays.toString(redOrder)+" ");
			
			// 1.2. BlueTeam: Random
			// - each member target one opponent
			// - randomly select target, then
			// - do basic sanity check: within range? etc.
			if (i%6==0) {
				blueOrder = targetSelectRand();
			}

			debugMessage(1,"blue "+java.util.Arrays.toString(blueOrder)+"   ");

			//int d1[] = {0, 1, 2, 3, 4};
			//int d2[] = {0, 1, 2, 3, 4};
			plotLineOfSight(redOrder,blueOrder,255-i*7);
			//plotLineOfSight(d1,d2,255-i*6);
			
		   // 2. Run stochastic sim with each team's selected targets
		   	runSimOnce(redOrder,blueOrder);
			if (dispFlag) {
				tsd.clearList();
			}
			plotObj();
			debugMessage(1,"e "+teamRed[0].e+", "+teamBlue[0].e+"\n");
		   	//curEnergySum *=0.9;
			//Thread.currentThread().sleep(1000);

		}

		// 3. Calculate outcome and assign winner
		
		//Thread.currentThread().sleep(300);
		if (dispFlag) {
			tsd.clearLineList();
			tsd.clearList();
		}
		return analyzeSim();
		//return rand.nextInt(3)-1; // or 1
	}
	catch ( Throwable e ) {
		logger.warn( "error evaluating chromosome "+c.toString(),e );
		return 0;
	}
}


/**
 * @see org.jgap.BulkFitnessFunction#getMaxFitnessValue()
 */
public int getMaxFitnessValue() {
	// YC: this function tells the main program what is the maximum
	// fitness attainable.
	//debugMessage(1,"maxFitness********"+((populationSize-1)*2)+"\n");
	if (randomOpponentFlag) {
		return (numTrials+numTrials);
	} else {
		return ((populationSize-1)*2*numTrials);
	}
}

} // end of class


